// -----------------------------------------------------------------------
// Litmus - DKFZ SIDT unit testing library
//
// Copyright (c) German Cancer Research Center (DKFZ),
// Software development for Integrated Diagnostic and Therapy (SIDT).
// ALL RIGHTS RESERVED.
// See litmusCopyright.txt or
// http://www.dkfz.de/en/sidt/projects/litmus/copyright.html
//
// This software is distributed WITHOUT ANY WARRANTY; without even
// the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
// PURPOSE.  See the above copyright notices for more information.
//
//------------------------------------------------------------------------
/*!
// @file
// @version $Revision: 4433 $ (last changed revision)
// @date    $Date: 2013-01-22 14:38:20 +0100 (Di, 22 Jan 2013) $ (last change date)
// @author  $Author: floca $ (last changed by)
*/


#ifndef __LIT_IMAGE_2TYPES_TESTER_TPP
#define __LIT_IMAGE_2TYPES_TESTER_TPP

#include "litTestImageIO.h"

#include "itkImageRegionConstIterator.h"
#include "itkAbsoluteValueDifferenceImageFilter.h"
#include "itkRescaleIntensityImageFilter.h"
#include <assert.h>

namespace lit 
{

	template<class TImage, class TTestImages>
	void
	  Image2TypesTester<TImage, TTestImages>::
	  setTestFileName(const StringType& testFileName)
	{
	  _testFileName = testFileName;
	};

	template<class TImage, class TTestImages>
	const StringType&
	  Image2TypesTester<TImage, TTestImages>::
	  getTestFileName() const
	{
	  return _testFileName;
	};

	template<class TImage, class TTestImages>
	void
	  Image2TypesTester<TImage, TTestImages>::
	  setSaveTestFiles(bool saveFiles)
	{
	  _saveTestFiles = saveFiles;
	};

	template<class TImage, class TTestImages>
	bool
	  Image2TypesTester<TImage, TTestImages>::
	  getSaveTestFiles() const
	{
	  return _saveTestFiles;
	};

	template<class TImage, class TTestImages>
	StringType
	  Image2TypesTester<TImage, TTestImages>::
	  getTestDescription(void) const
	{
	  return "Checks if actual image equals expected image";
	};

	template<class TImage, class TTestImages>
	StringType
	  Image2TypesTester<TImage, TTestImages>::
	  getTestName(void) const
	{
	  OStringStream stream;
	  stream << "Check equal: actual image == expected vector image (tolerance: +-"<<_checkThreshold<<")";

	  return stream.str();
	};

	template<class TImage, class TTestImages>
	void
	  Image2TypesTester<TImage, TTestImages>::
	  setExpectedImage( const ImageType* pImage)
	{
	  _spExpectedImage = pImage;
	};

	template<class TImage, class TTestImages>
	void
	  Image2TypesTester<TImage, TTestImages>::
	  setActualImage(const TestImageType* pImage)
	{
	  _spActualImage = pImage;
	};

	template<class TImage, class TTestImages>
	void
	  Image2TypesTester<TImage, TTestImages>::
	  setCheckThreshold(double checkThreshold)
	{
	  _checkThreshold = checkThreshold;
	};

	template<class TImage, class TTestImages>
	const typename Image2TypesTester<TImage, TTestImages>::ImageType*
	  Image2TypesTester<TImage, TTestImages>::
	  getExpectedImage() const
	{
	  return _spExpectedImage;
	};

	template<class TImage, class TTestImages>
	const typename Image2TypesTester<TImage, TTestImages>::TestImageType*
	  Image2TypesTester<TImage, TTestImages>::
	  getActualImage() const
	{
	  return _spActualImage;
	};


	template<class TImage, class TTestImages>
	double
	  Image2TypesTester<TImage, TTestImages>::
	  getCheckThreshold() const
	{
	  return _checkThreshold;
	};

	template<class TImage, class TTestImages>
	bool
	  Image2TypesTester<TImage, TTestImages>::
	  getCheckImageGeometry() const
	{
	  return _checkGeometry;
	};

	template<class TImage, class TTestImages>
	void
	  Image2TypesTester<TImage, TTestImages>::
	  setCheckImageGeometry(bool checkThreshold)
	{
	  _checkGeometry = checkThreshold;
	};

	template<class TImage, class TTestImages>
	Image2TypesTester<TImage, TTestImages>::
	  Image2TypesTester() : _checkThreshold(0.0), _testFileName(""), _saveTestFiles(false), _checkGeometry(true)
	{
	};

	template<class TImage, class TTestImages>
	Image2TypesTester<TImage, TTestImages>::
	  ~Image2TypesTester()
	{
	};

	template<class TImage, class TTestImages>
	bool
	  Image2TypesTester<TImage, TTestImages>::
	  doCheck(void) const
	{
	  bool imageEqualsImage = true;
	  _maxError = itk::NumericTraits<typename ImageType::PixelType>::Zero;
	  _errorPixelCount = 0;
	  _maxErrorIndex.Fill(0);

	  _pResults->onTestStart(getCurrentTestLabel());

	  if (_checkGeometry)
	  {
		if (!(_spExpectedImage->GetSpacing() == _spActualImage->GetSpacing()))
		{
		  imageEqualsImage = false;
		}
		if (!(_spExpectedImage->GetOrigin() == _spActualImage->GetOrigin()))
		{
		  imageEqualsImage = false;
		}
		if (!(_spExpectedImage->GetDirection() == _spActualImage->GetDirection()))
		{
		  imageEqualsImage = false;
		}
	  }

	  if (_spExpectedImage->GetLargestPossibleRegion() == _spActualImage->GetLargestPossibleRegion())
	  {
		//check for every Image pixel if it maps like the transform model.
		typedef itk::ImageRegionConstIterator< ImageType > ExpectedIteratorType;
		typedef itk::ImageRegionConstIterator< TestImageType > ActualIteratorType;
		ExpectedIteratorType expectedIterator(_spExpectedImage,_spExpectedImage->GetLargestPossibleRegion());
		ActualIteratorType actualIterator(_spActualImage,_spActualImage->GetLargestPossibleRegion());
		actualIterator.GoToBegin();
		for (expectedIterator.GoToBegin(); !(expectedIterator.IsAtEnd()); ++expectedIterator, ++actualIterator)
		{
		  typename ImageType::PixelType difference = abs(expectedIterator.Get() - actualIterator.Get());

		  if (difference>_checkThreshold)
		  {
			imageEqualsImage = false;
			++_errorPixelCount;
			if (_maxError<difference)
			{
			  _maxError = difference;
			  _maxErrorIndex = expectedIterator.GetIndex();
			};
		  }
		}
	  }
	  else
	  {
		imageEqualsImage = false;
	  }

	  return imageEqualsImage;
	};

	template<class TImage, class TTestImages>
	void
	  Image2TypesTester<TImage, TTestImages>::
	  handleSuccess(void) const
	{
	  OStringStream stream;
	  stream << "Pixels of "<<_spActualImage << std::endl <<"equals (tolerance: "<<_checkThreshold<<")"<<std::endl<<std::endl<< _spExpectedImage;

	  _pResults->onTestSuccess(getCurrentTestLabel(), stream.str());
	};

	template<class TImage, class TTestImages>
	void
	  Image2TypesTester<TImage, TTestImages>::
	  handleFailure(void) const
	{
	  OStringStream stream;
	  stream << "Pixels of "<<_spActualImage << std::endl << " doesn't equal (tolerance: "<<_checkThreshold<<") "<<std::endl<<std::endl<< _spExpectedImage << std::endl;
	  if (!(_spExpectedImage->GetLargestPossibleRegion() == _spActualImage->GetLargestPossibleRegion()))
	  {
		stream << std::endl << "Largest possible regions of images are not equal." << std::endl;
	  }
	  stream << std::endl << "Error pixel count:   "<< _errorPixelCount << std::endl;
	  stream << std::endl << "Maximum difference:  "<< _maxError << std::endl;
	  stream << std::endl << "Error pixel index:   "<< _maxErrorIndex << std::endl;
	  stream << std::endl << "<DartMeasurement name=\"Error pixel count\" type=\"numeric/int\">"<< _errorPixelCount << "</DartMeasurement>";
	  stream << std::endl << "<DartMeasurement name=\"Maximium error (difference)\" type=\"numeric/double\">"<< _maxError << "</DartMeasurement>";
	  stream << std::endl << "<DartMeasurement name=\"Error pixel index\" type=\"string\">"<< _maxErrorIndex << "</DartMeasurement>" << std::endl;

	  if (_saveTestFiles)
	  {
		saveTestFiles(stream);
	  }

	  _pResults->onTestFailure(getCurrentTestLabel(), stream.str());
	};

	template<class TImage, class TTestImages>
	void
	  Image2TypesTester<TImage, TTestImages>::
	  saveTestFiles(OStringStream& stream) const
	{
	  typedef itk::AbsoluteValueDifferenceImageFilter<ImageType,TestImageType,TestImageType> AbsoluteDiffType;
	  typedef itk::RescaleIntensityImageFilter<TestImageType,TestImageType> RescaleType;

	  typedef TestImageIO<typename TestImageType::PixelType, TestImageType> ActualTestIOType;
	  typedef TestImageIO<typename ImageType::PixelType, ImageType> ExpectedTestIOType;

	  if (_spExpectedImage->GetLargestPossibleRegion() == _spActualImage->GetLargestPossibleRegion())
	  { //the AbsoluteValueDifferenceImageFilter crashes if the regions are unequal
		//difference image
		typename AbsoluteDiffType::Pointer spDiffFilter = AbsoluteDiffType::New();
		spDiffFilter->SetInput1(_spExpectedImage);
		spDiffFilter->SetInput2(_spActualImage);

		typename RescaleType::Pointer spRescaleFilter = RescaleType::New();
		spRescaleFilter->SetOutputMinimum(itk::NumericTraits<typename TestImageType::PixelType>::NonpositiveMin());
		spRescaleFilter->SetOutputMaximum(itk::NumericTraits<typename TestImageType::PixelType>::max());
		spRescaleFilter->SetInput(spDiffFilter->GetOutput());

		OStringStream diffName;
		if (TestImageType::ImageDimension==2)
		{
		  diffName << _testFileName << ".diff.png";
		}
		else
		{
		  diffName << _testFileName << ".diff.mhd";
		}

		OStringStream normDiffName;
		if (TestImageType::ImageDimension==2)
		{
		  normDiffName << _testFileName << ".diff.norm.png";
		}
		else
		{
		  normDiffName << _testFileName << ".diff.norm.mhd";
		}

		try
		{
		  spRescaleFilter->Update();
		}
		catch (...)
		{
		  std::cerr << "Error during rescale of " << diffName.str() << std::endl;
		}

		try
		{
		  ActualTestIOType::writeImage(spDiffFilter->GetOutput(),diffName.str());
		}
		catch (...)
		{
		  std::cerr << "Error during write of " << diffName.str() << std::endl;
		}

		try
		{
		  ActualTestIOType::writeImage(spRescaleFilter->GetOutput(),normDiffName.str());
		}
		catch (...)
		{
		  std::cerr << "Error during write of " << normDiffName.str() << std::endl;
		}

		if (TestImageType::ImageDimension==2)
		{
		  stream << "<DartMeasurementFile name=\"DifferenceImage\" type=\"image/png\">";
		}
		else
		{
		  stream << "<DartMeasurementFile name=\"DifferenceImage\" type=\"image/mhd\">";
		}
		stream << diffName.str();
		stream << "</DartMeasurementFile>" << std::endl;

		if (TestImageType::ImageDimension==2)
		{
		  stream << "<DartMeasurementFile name=\"NormalizedDifferenceImage\" type=\"image/png\">";
		}
		else
		{
		  stream << "<DartMeasurementFile name=\"NormalizedDifferenceImage\" type=\"image/mhd\">";
		}
		stream << normDiffName.str();
		stream << "</DartMeasurementFile>" << std::endl;
	  }

	  //expected image output
	  OStringStream expectedName;
	  if (TestImageType::ImageDimension==2)
	  {
		expectedName << _testFileName << ".exp.png";
	  }
	  else
	  {
		expectedName << _testFileName << ".exp.mhd";
	  }

	  try
	  {
		ExpectedTestIOType::writeImage(_spExpectedImage,expectedName.str());
	  }
	  catch (...)
	  {
		std::cerr << "Error during write of " << expectedName.str() << std::endl;
	  }

	  if (TestImageType::ImageDimension==2)
	  {
		stream << "<DartMeasurementFile name=\"ExpectedImage\" type=\"image/png\">";
	  }
	  else
	  {
		stream << "<DartMeasurementFile name=\"ExpectedImage\" type=\"image/mhd\">";
	  }
	  stream << expectedName.str();
	  stream << "</DartMeasurementFile>" << std::endl;

	  //actual image output
	  OStringStream actualName;
	  if (TestImageType::ImageDimension==2)
	  {
		actualName << _testFileName << ".act.png";
	  }
	  else
	  {
		actualName << _testFileName << ".act.mhd";
	  }

	  try
	  {
		ActualTestIOType::writeImage(_spActualImage,actualName.str());
	  }
	  catch (...)
	  {
		std::cerr << "Error during write of " << actualName.str() << std::endl;
	  }

	  if (TestImageType::ImageDimension==2)
	  {
		stream << "<DartMeasurementFile name=\"ActualImage\" type=\"image/png\">";
	  }
	  else
	  {
		stream << "<DartMeasurementFile name=\"ActualImage\" type=\"image/mhd\">";
	  }
	  stream << actualName.str();
	  stream << "</DartMeasurementFile>" << std::endl;
	};

}

#endif
