// -----------------------------------------------------------------------
// Litmus - DKFZ SIDT unit testing library
//
// Copyright (c) German Cancer Research Center (DKFZ),
// Software development for Integrated Diagnostic and Therapy (SIDT).
// ALL RIGHTS RESERVED.
// See litmusCopyright.txt or
// http://www.dkfz.de/en/sidt/projects/litmus/copyright.html
//
// This software is distributed WITHOUT ANY WARRANTY; without even
// the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
// PURPOSE.  See the above copyright notices for more information.
//
//------------------------------------------------------------------------
/*!
// @file
// @version $Revision: 4433 $ (last changed revision)
// @date    $Date: 2013-01-22 14:38:20 +0100 (Di, 22 Jan 2013) $ (last change date)
// @author  $Author: floca $ (last changed by)
*/


#ifndef __LIT_TRANSFORM_FIELD_TESTER_TPP
#define __LIT_TRANSFORM_FIELD_TESTER_TPP

#include "itkImageRegionConstIterator.h"
#include <assert.h>

namespace lit 
{

    template<class TField, class TTransform>
    StringType
      TransformFieldTester<TField,TTransform>::
      getTestDescription(void) const
    {
      return "Checks if actual deformation field equals given transformation model";
    };

    template<class TField, class TTransform>
    StringType
      TransformFieldTester<TField,TTransform>::
      getTestName(void) const
    {
      OStringStream stream;
      stream << "Check equal: deformation field == transformation model (tolerance: +-"<<_checkThreshold<<")";

      return stream.str();
    };

    template<class TField, class TTransform>
    void
      TransformFieldTester<TField,TTransform>::
      setReferenceTransform(const TransformType* pTransform)
    {
      _spReferenceTransform = pTransform;
    };

    template<class TField, class TTransform>
    void
      TransformFieldTester<TField,TTransform>::
      setActualField(const FieldType* pField)
    {
      _spActualField = pField;
    };

    template<class TField, class TTransform>
    void
      TransformFieldTester<TField,TTransform>::
      setCheckThreshold(double checkThreshold)
    {
      _checkThreshold = checkThreshold;
    };

    template<class TField, class TTransform>
    const typename TransformFieldTester<TField,TTransform>::TransformType*
      TransformFieldTester<TField,TTransform>::
      getReferenceTransform() const
    {
      return _spReferenceTransform;
    };

    template<class TField, class TTransform>
    const typename TransformFieldTester<TField,TTransform>::FieldType*
      TransformFieldTester<TField,TTransform>::
      getActualField() const
    {
      return _spActualField;
    };


    template<class TField, class TTransform>
    double
      TransformFieldTester<TField,TTransform>::
      getCheckThreshold() const
    {
      return _checkThreshold;
    };

    template<class TField, class TTransform>
    void
      TransformFieldTester<TField,TTransform>::
      setTestRegion(const TestRegionType& testRegion)
    {
      _testRegion = testRegion;
      _useTestRegion = true;
    };
      
      template<class TField, class TTransform>
    const typename TransformFieldTester<TField,TTransform>::TestRegionType&
      TransformFieldTester<TField,TTransform>::
       getTestRegion() const
    {
      return _testRegion;
    };

    template<class TField, class TTransform>
    TransformFieldTester<TField,TTransform>::
      TransformFieldTester() :  _useTestRegion(false), _checkThreshold(0.0)
    {
    };

    template<class TField, class TTransform>
    TransformFieldTester<TField,TTransform>::
      ~TransformFieldTester()
    {
    };

    /*! performes the test and checks the results.
    * @result Indicates if the test was successfull (true) or if it failed (false)
    */
    template<class TField, class TTransform>
    bool
      TransformFieldTester<TField,TTransform>::
      doCheck(void) const
    {
      bool fieldEqualsModel = true;
      _maxError = 0.0;
      _errorPixelCount = 0;
      _maxErrorIndex.Fill(0);

      _pResults->onTestStart(getCurrentTestLabel());

      //check for every field pixel if it lits like the transform model.
      typedef itk::ImageRegionConstIterator< FieldType > IteratorType;

      TestRegionType currentTestRegion = _testRegion;
      if (!_useTestRegion)
      {
        currentTestRegion = _spActualField->GetLargestPossibleRegion();
      }
      IteratorType iterator(_spActualField,currentTestRegion);
      for (iterator.GoToBegin(); !(iterator.IsAtEnd()); ++iterator)
      {
        typename FieldType::IndexType index = iterator.GetIndex();
        typename FieldType::PointType inPoint;
        _spActualField->TransformIndexToPhysicalPoint(index,inPoint);

        typename TransformType::OutputPointType transformOutPoint = _spReferenceTransform->TransformPoint(inPoint);
        typename IteratorType::PixelType fieldvector = iterator.Get();
        typename TransformType::OutputPointType fieldOutPoint = inPoint + iterator.Get();
        typename TransformType::OutputVectorType differenceVector = transformOutPoint-fieldOutPoint;
        double difference = differenceVector.GetNorm();

        if (difference>_checkThreshold)
        {
          fieldEqualsModel = false;
          ++_errorPixelCount;
          if (_maxError<difference)
          {
            _maxError = difference;
            _maxErrorIndex = index;
          };
        }
      }

      return fieldEqualsModel;
    };

    template<class TField, class TTransform>
    void
      TransformFieldTester<TField,TTransform>::
      handleSuccess(void) const
    {
      OStringStream stream;
      stream << "Vectors of "<<_spActualField << std::endl <<"equals (tolerance: "<<_checkThreshold<<")"<<std::endl<<std::endl<< _spReferenceTransform;
      _pResults->onTestSuccess(getCurrentTestLabel(), stream.str());
    };

    template<class TField, class TTransform>
    void
      TransformFieldTester<TField,TTransform>::
      handleFailure(void) const
    {
      OStringStream stream;
      stream << "Vectors of "<<_spActualField << std::endl << " doesn't equal (tolerance: "<<_checkThreshold<<") "<<std::endl<<std::endl<< _spReferenceTransform << std::endl;
      stream << std::endl << "Error pixel count:   "<< _errorPixelCount << std::endl;
      stream << std::endl << "Maximum difference:  "<< _maxError << std::endl;
      stream << std::endl << "Error pixel index:   "<< _maxErrorIndex << std::endl;
      stream << std::endl << "<DartMeasurement name=\"Error pixel count\" type=\"numeric/int\">"<< _errorPixelCount << "</DartMeasurement>";
      stream << std::endl << "<DartMeasurement name=\"Maximium error (difference)\" type=\"numeric/double\">"<< _maxError << "</DartMeasurement>";
      stream << std::endl << "<DartMeasurement name=\"Error pixel index\" type=\"string\">"<< _maxErrorIndex << "</DartMeasurement>" << std::endl;

      _pResults->onTestFailure(getCurrentTestLabel(), stream.str());
    };

}

#endif
